package org.benchkit;

import java.lang.management.CompilationMXBean;
import java.lang.management.ManagementFactory;
import java.util.List;

/**
 * A {@link BenchmarkRunner} object is in charge of actually running a
 * benchmark; making sure the code is properly warmed up by the JIT,
 * and ensuring that the results are reported.
 * @author cvh
 *
 */
public class BenchmarkRunner {
  private static final CompilationMXBean COMPILER =
      ManagementFactory.getCompilationMXBean();
  
  public static final int DEFAULT_WARMUP_ITERATIONS =
      Math.max(1, Integer.getInteger("benchkit.iterations.warmup", 10));
  
  public static final int DEFAULT_ITERATIONS =
      Math.max(1, Integer.getInteger("benchkit.iterations", 10));
  
  static {
    if (!COMPILER.isCompilationTimeMonitoringSupported()) {
      System.err.println(
          "Warning: JIT compilation time monitoring through JMX is not " +
          "supported on this JVM. Benchmark sessions with noticible JIT " +
          "compile lags will not be discarded from the recordings.");
    }
  }
  
  private Benchmark benchmark;
  private Reporter reporter = new PrintingReporter();
  private Reporter warmupReporter = new WarmupPrintingReporter();
  private Reporter combinedReporter = null;
  private int warmupIterations = DEFAULT_WARMUP_ITERATIONS;
  private int iterations = DEFAULT_ITERATIONS;
  private Object[] arguments;
  
  /**
   * Create a runner for the given benchmark.
   * @param benchmark The benchmark to run.
   * @param arguments The list of arguments given to the benchmark. Used for
   * reporting the benchmarks particular configuration.
   */
  public BenchmarkRunner(Benchmark benchmark, Object... arguments) {
    this.benchmark = benchmark;
    this.arguments = arguments;
  }

  /**
   * The minimum number of benchmark sessions the runner will use for warm-up.
   * @return By default 10, or the value of the
   * {@code benchkit.iterations.warmup} system property.
   */
  public int getWarmupIterations() {
    return warmupIterations;
  }

  /**
   * Set the minimum number of benchmark sessions the runner will use for
   * warm-up.
   * 
   * The default is 10, but this can be overwritten by setting the
   * {@code benchkit.iterations.warmup} system property.
   * @param warmupIterations
   */
  public void setWarmupIterations(int warmupIterations) {
    if (warmupIterations < 0) {
      throw new IllegalArgumentException(
          "warmupIterations cannot be less than 0.");
    }
    this.warmupIterations = warmupIterations;
  }

  /**
   * The number of benchmark sessions the runner will record and report.
   * @return By default 10, or the value of the {@code benchkit.iterations}
   * system property.
   */
  public int getIterations() {
    return iterations;
  }

  /**
   * Set the number of benchmark sessions the runner will record and report.
   * 
   * The default is 10, but this can be overwritten by setting the
   * {@code benchkit.iterations} system property.
   * @param iterations
   */
  public void setIterations(int iterations) {
    if (iterations < 1) {
      throw new IllegalArgumentException("iterations cannot be less than 1.");
    }
    this.iterations = iterations;
  }

  /**
   * Get the {@link Reporter} that will be used to report the results of the
   * benchmark sessions.
   * @return The configured Reporter. Never null.
   */
  public Reporter getReporter() {
    return reporter;
  }

  /**
   * Set the reporter that will be used to report the results of the benchmark
   * sessions, as well as the combined result for all of the benchmark sessions.
   * @param reporter
   */
  public void setReporter(Reporter reporter) {
    if (reporter == null) {
      throw new IllegalArgumentException("reporter cannot be null.");
    }
    this.reporter = reporter;
  }
  
  /**
   * The reporter used for warm-up iterations.
   */
  public Reporter getWarmupReporter() {
    return warmupReporter;
  }

  /**
   * Set the reporter used for warm-up iterations.
   * @param warmupReporter
   */
  public void setWarmupReporter(Reporter warmupReporter) {
    this.warmupReporter = warmupReporter;
  }

  /**
   * The reporter used for reporting combined results, if any.
   */
  public Reporter getCombinedReporter() {
    return combinedReporter;
  }

  /**
   * Set the reporter used for reporting combined results. If it is
   * <code>null</code>, which is the default, then the
   * {@link #getReporter() regular reporter} is used instead.
   * @param combinedReporter
   */
  public void setCombinedReporter(Reporter combinedReporter) {
    this.combinedReporter = combinedReporter;
  }

  /**
   * Actually run the benchmark, once it has been configured.
   * @throws Exception
   */
  public void run() throws Exception {
    benchmark.setUp();
    
    String benchmarkName =
        benchmark.getClass().getSimpleName();
    Recorder recorder = new Recorder(benchmarkName);
    Recorder combinedRecorder = new Recorder("Combined Results: " + benchmarkName);
    
    int iterationsLeft = iterations;
    int warmupIterationsLeft = warmupIterations;
    do {
      reportArguments();
      long jitTimeStart = COMPILER.getTotalCompilationTime();
      System.gc();
      recorder.startSession();
      benchmark.runSession(recorder);
      recorder.endSession();
      long jitTime = COMPILER.getTotalCompilationTime() - jitTimeStart;
      
      if (jitTime == 0 && warmupIterationsLeft <= 0) {
        iterationsLeft--;
        recorder.report(reporter);
        combinedRecorder.add(recorder);
      } else {
        warmupIterationsLeft--;
        // We report to the warmupReporter, because reporting itself might
        // end up spending JIT time.
        // In particular, the string formatting is very likely to spend time
        // getting JIT compiled.
        recorder.report(warmupReporter);
        warmupReporter.clear();
        System.err.printf(
            "Warm-up round. (JIT spent %s ms. of CPU time)\n", jitTime);
      }
      recorder.reset();
      reporter.clear();
    } while (iterationsLeft > 0);
    
    benchmark.tearDown();
    Reporter finalReporter =
        combinedReporter == null? reporter : combinedReporter;
    finalReporter.setArguments(arguments);
    combinedRecorder.report(finalReporter);
    finalReporter.clear();
    System.out.println("## Done.");
  }

  protected void reportArguments() {
    reporter.setArguments(arguments);
    warmupReporter.setArguments(arguments);
    if (combinedReporter != null) {
      combinedReporter.setArguments(arguments);
    }
  }
  
  /**
   * Run the given benchmark with a default {@link BenchmarkRunner}
   * configuration.
   * @param benchmark The {@link Benchmark} to run.
   * @param arguments The list of arguments given to the benchmark. Used for
   * reporting the benchmarks particular configuration.
   * @throws Exception Any exception the benchmark might throw.
   */
  public static void run(Benchmark benchmark, Object... arguments)
      throws Exception {
    run(new PrintingReporter(), benchmark, arguments);
  }
  
  /**
   * Run the given benchmark with a default {@link BenchmarkRunner}
   * configuration.
   * @param reporter The {@link Reporter} used for reporting the results of
   * running the benchmark.
   * @param benchmark The {@link Benchmark} to run.
   * @param arguments The list of arguments given to the benchmark. Used for
   * reporting the benchmarks particular configuration.
   * @throws Exception Any exception the benchmark might throw.
   */
  public static void run(
      Reporter reporter,
      Benchmark benchmark,
      Object... arguments)
      throws Exception {
    BenchmarkRunner runner = new BenchmarkRunner(benchmark, arguments);
    runner.setReporter(reporter);
    runner.run();
  }
  
  /**
   * Run all configured permutations of the provided benchmark type.
   * 
   * Instances are created using parameter injection with the {@link Param}
   * annotation.
   * @param benchmarkType The type of benchmarks to run. Constructor parameters
   * must be annotated with {@link Param}.
   * @throws Exception
   */
  public static void run(Class<? extends Benchmark> benchmarkType) throws Exception {
    run(benchmarkType, new PrintingReporter());
  }
  
  /**
   * Run all configured permutations of the provided benchmark type.
   * 
   * Instances are created using parameter injection with the {@link Param}
   * annotation.
   * @param benchmarkType The type of benchmarks to run. Constructor parameters
   * must be annotated with {@link Param}.
   * @param reporter The {@link Reporter} used to report the results of running
   * the benchmarks.
   * @throws Exception
   */
  public static void run(
      Class<? extends Benchmark> benchmarkType,
      Reporter reporter) throws Exception {
    run(benchmarkType, reporter, DEFAULT_ITERATIONS, DEFAULT_WARMUP_ITERATIONS);
  }
  
  public static void run(
      Class<? extends Benchmark> benchmarkType,
      Reporter reporter,
      int iterationCount,
      int warmupIterationCount) throws Exception {
    run(benchmarkType, reporter, reporter, iterationCount, warmupIterationCount);
  }
  
  public static void run(
      Class<? extends Benchmark> benchmarkType,
      Reporter reporter,
      Reporter combinedReporter,
      int iterationCount,
      int warmupIterationCount) throws Exception {
    List<BenchmarkRunner> runners =
        BenchmarkRunnerGenerator.createRunners(benchmarkType);
    for (BenchmarkRunner runner : runners) {
      runner.setReporter(reporter);
      runner.setCombinedReporter(combinedReporter);
      runner.setIterations(iterationCount);
      runner.setWarmupIterations(warmupIterationCount);
      runner.run();
    }
  }
}
