package org.benchkit;

import java.util.concurrent.TimeUnit;

import org.HdrHistogram.Histogram;

/**
 * A container for recorded values during a benchmark.
 * 
 * Note that this class is NOT thread-safe. A benchmark recording from more
 * than one thread, must give each thread a recorder, and then combine them
 * afterwards with {@link #add(Recorder)}.
 * 
 * @author cvh
 */
public final class Recorder {
  
  private static final long DEFAULT_HIGHEST_EXPECTED_LATENCY =
      TimeUnit.SECONDS.toNanos(60);
  private static final double MILLIS_TO_NANOS = TimeUnit.MILLISECONDS.toNanos(1);
  
  private String benchmarkName;
  private Histogram histogram;
  private long sessionStart;
  private long sessionEnd;
  // We maintain our own count of the total recordings because the histogram
  // might overflow and forget some:
  private long totalRecordings;
  private long compositeTime;
  // We maintain our own min and max values, because the histogram is lossy
  private long min = Long.MAX_VALUE;
  private long max = Long.MIN_VALUE;
  
  /**
   * Create a recorder with a given benchmark name. The benchmark name is
   * used when the recorded values are reported to a {@link Reporter}.
   * 
   * The recorder is configured to expect a highest measured latency value of
   * no larger than 1 minute.
   * 
   * @param benchmarkName The name of the benchmark for which this recorder is
   * used.
   */
  public Recorder(String benchmarkName) {
    this(benchmarkName, DEFAULT_HIGHEST_EXPECTED_LATENCY);
  }

  /**
   * Create a recorder with a given benchmark name, and a given expected
   * latency ceiling in nanoseconds.
   * @param benchmarkName The name of the benchmark for which this recorder is
   * used.
   * @param highestExpectedLatencyNanos The maximum latency that one is
   * expecting to record with this recorder, measured in nanoseconds.
   * A minute or more is a good default, as the internal representation is
   * very space efficient.
   */
  public Recorder(String benchmarkName, long highestExpectedLatencyNanos) {
    this.benchmarkName = benchmarkName;
    histogram = new Histogram(highestExpectedLatencyNanos, 4);
  }
  
  /**
   * Create a new recorder that is configured just like this one, but contains
   * no data.
   * 
   * This method is useful if you have a benchmark that spawns a number of
   * threads, and you want to give each of them their own recorder: just
   * create blank copies of the "main recorder" for the benchmark.
   * Just remember to {@link #add(Recorder) add} the thread-local recorders
   * back into the main recorder, once the benchmark session is done.
   * @return a new Recorder, just like this one, but without any data.
   */
  public Recorder createBlankCopy() {
    return new Recorder(benchmarkName, histogram.getHighestTrackableValue());
  }
  
  /**
   * Called at the beginning of a benchmarking session.
   * 
   * Used to measure the elapsed running time.
   */
  void startSession() {
    sessionStart = System.currentTimeMillis();
  }
  
  /**
   * Called at the end of a benchmarking session.
   * 
   * Used to measure the elapsed running time.
   */
  void endSession() {
    sessionEnd = System.currentTimeMillis();
  }
  
  /**
   * Called just prior to the individual operation. This will increment the
   * "event" or "transaction" count, and return a value that is passed to
   * {@link #record(long)} for measuring the latency.
   * @return a value that must be passed to {@link #record(long)} after
   * executing the block of code for which we are measuring latency.
   */
  public long begin() {
    return System.nanoTime();
  }
  
  /**
   * Record that an event or transaction has transpired, giving the value
   * returned from the last call to {@link #begin()}.
   * @param beginValue
   * @return A new 'begin' value, for those tight loops.
   */
  public long record(long beginValue) {
    long now = begin();
    long value = now - beginValue;
    try {
      histogram.recordValue(value);
    } catch (RuntimeException e) {
      err("## Histogram has overflowed: " + histogram.hasOverflowed());
      err("## Highest expected value: " + histogram.getHighestTrackableValue());
      err("## Tried to record value: " + value);
      throw e;
    }
    totalRecordings++;
    min = Math.min(min, value);
    max = Math.max(max, value);
    return now;
  }
  
  private static void err(String msg) {
    System.err.println(msg);
  }
  
  private long getCompositeTime() {
    return compositeTime + (sessionEnd - sessionStart);
  }
  
  /**
   * Add the data collected by the given recorder, to this recorder.
   * @param recorder The recorder whose data should be added to this recorder.
   */
  public void add(Recorder recorder) {
    if (benchmarkName == null) {
      benchmarkName = recorder.benchmarkName;
    }
    
    compositeTime += recorder.getCompositeTime();
    totalRecordings += recorder.totalRecordings;
    min = Math.min(min, recorder.min);
    max = Math.max(max, recorder.max);
    histogram.add(recorder.histogram);
  }
  
  /**
   * Report the recorded results to the given {@link Reporter}.
   * @param reporter
   */
  void report(Reporter reporter) {
    long elapsedRuntimeMillis = getCompositeTime();
    
    RecordingData data = new RecordingData(
        benchmarkName,
        elapsedRuntimeMillis,
        totalRecordings,
        histogram.copy(),
        MILLIS_TO_NANOS,
        min,
        max);
    
    reporter.report(data);
  }
  
  /**
   * Resets the recorded values in the Recorder, but maintains the settings
   * given through the constructor.
   * 
   * This method allows Recorders to be reused between sessions.
   */
  void reset() {
    sessionStart = 0;
    sessionEnd = 0;
    totalRecordings = 0;
    min = Long.MAX_VALUE;
    max = Long.MIN_VALUE;
    histogram.reset();
  }
}
