package org.benchkit.htmlchartsreporter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public final class ChartRenderUtil {

  public interface CurveStrategy {
    double yvalue(List<DataPoint> points);
  }

  public interface CurveDecorator {
    void decorate(Map<String, Object> curve, Map<String, Object> linesConfig);
  }

  public static final CurveStrategy throughputMean = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      long opssecSum = 0;
      for (DataPoint point : points) {
        opssecSum += point.opssec;
      }
      double opssec = (double) opssecSum / points.size();
      return opssec;
    }
  };
  
  public static final CurveStrategy latencyMean = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      long latencySum = 0;
      for (DataPoint point : points) {
        latencySum += point.latencyMean;
      }
      double toMillis = points.get(0).factorConvertingLatencyToMillis;
      double latency = (latencySum / toMillis) / points.size();
      return latency;
    }
  };
  
  public static final CurveStrategy latencyDev = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      long latencySum = 0;
      for (DataPoint point : points) {
        latencySum += point.latencyStddev;
      }
      double toMillis = points.get(0).factorConvertingLatencyToMillis;
      double latency = (latencySum / toMillis) / points.size();
      return latency;
    }
  };
  
  public static final CurveStrategy latencyMax = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      double max = Double.MIN_VALUE;
      for (DataPoint point : points) {
        max = Math.max(max, point.latencyMax);
      }
      double toMillis = points.get(0).factorConvertingLatencyToMillis;
      return max / toMillis;
    }
  };
  
  public static final CurveStrategy latency9999 = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      long latencySum = 0;
      for (DataPoint point : points) {
        latencySum += point.latency9999;
      }
      double toMillis = points.get(0).factorConvertingLatencyToMillis;
      double latency = (latencySum / toMillis) / points.size();
      return latency;
    }
  };
  
  public static final CurveStrategy latency999 = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      long latencySum = 0;
      for (DataPoint point : points) {
        latencySum += point.latency999;
      }
      double toMillis = points.get(0).factorConvertingLatencyToMillis;
      double latency = (latencySum / toMillis) / points.size();
      return latency;
    }
  };
  
  public static final CurveStrategy latency99 = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      long latencySum = 0;
      for (DataPoint point : points) {
        latencySum += point.latency99;
      }
      double toMillis = points.get(0).factorConvertingLatencyToMillis;
      double latency = (latencySum / toMillis) / points.size();
      return latency;
    }
  };
  
  public static final CurveStrategy latency95 = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      long latencySum = 0;
      for (DataPoint point : points) {
        latencySum += point.latency95;
      }
      double toMillis = points.get(0).factorConvertingLatencyToMillis;
      double latency = (latencySum / toMillis) / points.size();
      return latency;
    }
  };
  
  public static final CurveStrategy latencyMin = new CurveStrategy() {
    public double yvalue(List<DataPoint> points) {
      double min = Double.MAX_VALUE;
      for (DataPoint point : points) {
        min = Math.min(min, point.latencyMin);
      }
      double toMillis = points.get(0).factorConvertingLatencyToMillis;
      return min / toMillis;
    }
  };
  
  public static final CurveDecorator noLineFill = new CurveDecorator() {
    public void decorate(
        Map<String, Object> curve,
        Map<String, Object> linesConfig) {
      linesConfig.put("lineWidth", 0);
      linesConfig.put("fill", false);
    }
  };
  
  public static final Comparator<DataPoint> byOpssec = new Comparator<DataPoint>() {
    public int compare(DataPoint a, DataPoint b) {
      return (int) (a.opssec - b.opssec);
    }
  };

  private ChartRenderUtil() {
    // Class is static only
  }
  
  public static Hsl[] generateColours() {
    Hsl[] colours = new Hsl[12];
    for (int i = 0; i < 12; i++) {
      colours[i] = new Hsl((i*150)%360, 75, 65);
    }
    return colours;
  }

  public static void render(Template tpl, String panelName, String xaxisName,
      String yaxisName, List<Object> dataset) {
    tpl.setReplacer("panel", panelName);
    tpl.setReplacer("dataset", dataset);
    tpl.setReplacer("xaxisName", xaxisName);
    tpl.setReplacer("yaxisName", yaxisName);
  }

  @SuppressWarnings("unchecked")
  public static void renderCurve(
      CurveStrategy strategy,
      List<Object> dataset,
      DataSeries dataSeries,
      Map<Comparable<?>, Integer> xvalueMap,
      Hsl hsl,
      CurveDecorator... decorators) {
    List<List<?>> dataPoints = new ArrayList<List<?>>();
    for (List<DataPoint> points : dataSeries) {
      Object xvalue = points.get(0).xvalue;
      double yvalue = strategy.yvalue(points);
      dataPoints.add(Arrays.asList(xvalueMap.get(xvalue), yvalue));
    }
    
    Map<String, Object> linesConfig = createLinesConfig();
    Map<String, Object> curve = createCurve(linesConfig, dataPoints, hsl);
    for (CurveDecorator decorator : decorators) {
      decorator.decorate(curve, linesConfig);
    }
    dataset.add(JsonOut.json(curve));
  }

  public static Map<String, Object> createLinesConfig() {
    Map<String, Object> linesConfig = new HashMap<String, Object>();
    linesConfig.put("show", true);
    return linesConfig;
  }

  public static Map<String, Object> createCurve(
      Map<String, Object> linesConfig,
      List<List<?>> dataPoints,
      Hsl hsl) {
    Map<String, Object> curve = new HashMap<String, Object>();
    curve.put("color", hsl.toCssHsl());
    curve.put("lines", linesConfig);
    curve.put("data", dataPoints);
    return curve;
  }

  public static CurveStrategy specificOpssec(final int index) {
    return new CurveStrategy() {
      public double yvalue(List<DataPoint> points) {
        return points.get(index).opssec;
      }
    };
  }

  public static CurveDecorator label(final String name) {
    return new CurveDecorator() {
      public void decorate(
          Map<String, Object> curve,
          Map<String, Object> linesConfig) {
        curve.put("label", name);
      }
    };
  }

  public static CurveDecorator id(final String id) {
    return new CurveDecorator() {
      public void decorate(
          Map<String, Object> curve,
          Map<String, Object> linesConfig) {
        curve.put("id", id);
      }
    };
  }

  public static CurveDecorator withNormalFillBetween(
      final int id,
      final Hsl colour,
      final int curveIndex,
      final int curveCount) {
    final double n = curveIndex;
    final double m = curveCount;
    return new CurveDecorator() {
      public void decorate(
          Map<String, Object> curve,
          Map<String, Object> linesConfig) {
  
        linesConfig.put("lineWidth", 0);
        if (curveIndex == 0) {
          linesConfig.put("fill", false);
        } else {
          double colourAlpha = 0.1 + (m/2 - Math.abs(n - m/2)) / m / 3;
          linesConfig.put("fill", true);
          linesConfig.put("fillColor", colour.toCssHsla(colourAlpha));
          curve.put("fillBetween", curveId(curveIndex - 1));
        }
        curve.put("id", curveId(curveIndex));
      }
  
      private String curveId(int i) {
        return "s" + id + "_" + i;
      }
    };
  }

  public static CurveDecorator fillBetween(
      final String theirId,
      final String colour) {
    return new CurveDecorator() {
      public void decorate(
          Map<String, Object> curve,
          Map<String, Object> linesConfig) {
        linesConfig.put("lineWidth", 0);
        linesConfig.put("fill", true);
        linesConfig.put("fillColor", colour);
        curve.put("fillBetween", theirId);
      }
    };
  }
}
