package org.benchkit.htmlchartsreporter;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import org.benchkit.RecordingData;
import org.benchkit.Reporter;

public class HtmlChartsReporter implements Reporter {
  private final DataInterpretor interpretor;
  private final String chartNameTemplate;
  private final Map<String, Chart> charts;
  private final List<ChartRender> chartRenders;
  private Object[] arguments;
  private int chartCounter = 0;

  public HtmlChartsReporter(DataInterpretor interpretor, String chartNameTemplate) {
    this.interpretor = interpretor;
    this.chartNameTemplate = chartNameTemplate;
    charts = new HashMap<String, Chart>();
    chartRenders = new LinkedList<ChartRender>();
  }
  
  public void addChartRender(ChartRender render) {
    chartRenders.add(render);
  }

  @Override
  public void setArguments(Object[] arguments) {
    this.arguments = arguments;
  }

  @Override
  public void report(RecordingData recordingData) {
    Chart chart = getOrCreateChart();
    DataPoint dataPoint = createDataPoint(recordingData);
    chart.addDataPoint(dataPoint);
  }

  private DataPoint createDataPoint(RecordingData data) {
    String seriesName = interpretor.getSeriesName(arguments);
    Comparable<Object> xvalue = interpretor.getXvalue(arguments);
    DataPoint dataPoint = new DataPoint(seriesName, xvalue, data);
    return dataPoint;
  }

  private Chart getOrCreateChart() {
    String chartName = String.format(chartNameTemplate, arguments);
    Chart chart = charts.get(chartName);
    if (chart == null) {
      chart = new Chart(++chartCounter);
      charts.put(chartName, chart);
    }
    return chart;
  }

  @Override
  public void clear() {
    arguments = null;
  }

  private static String readResource(String resource) { // what a bother...
    InputStream stream = HtmlChartsReporter.class.getResourceAsStream(resource);
    Scanner scanner = new Scanner(stream, "UTF-8").useDelimiter("\\A");
    return scanner.hasNext() ? scanner.next() : "";
  }

  public String generateReport() {
    for (ChartRender render : chartRenders) {
      render.calibrate(charts.values());
    }
    Template report = new Template(readResource("template.html"));
    report.setReplacer("BENCHMARK_NAME", interpretor.getBenchmarkName(arguments));
    report.setReplacer("JQUERY", readResource("jquery.js"));
    report.setReplacer("JQUERY_FLOT", readResource("jquery-flot.js"));
    report.setReplacer("JQUERY_FLOT_FILLBETWEEN", readResource("jquery-flot-fillbetween.js"));
    report.setReplacer("JQUERY_FLOT_AXISLABELS", readResource("jquery-flot-axislabels.js"));
    report.setReplacer("PANEL_DIVS", generatePanelDivs());
    report.setReplacer("CHART_DATA", generateChartData());
    return report.toString();
  }
  
  private Object generatePanelDivs() {
    List<String> divs = new ArrayList<String>();
    Template headlineTpl = new Template(readResource("panel-headline-template.txt"));
    Template panelTpl = new Template(readResource("panels-template.txt"));
    for (Map.Entry<String, Chart> chartEntry : charts.entrySet()) {
      String chartName = chartEntry.getKey();
      Chart chart = chartEntry.getValue();
      divs.add("<h3>" + chartName + "</h3>");
      for (ChartRender render : chartRenders) {
        render.renderPanels(chart, panelTpl, headlineTpl, divs);
      }
    }
    return new Template.ListReplacer(divs);
  }

  private Object generateChartData() {
    List<String> chartDataScripts = new ArrayList<String>();
    Template tpl = new Template(readResource("flotchart-template.txt"));
    for (Chart chart : charts.values()) {
      for (ChartRender render : chartRenders) {
        render.renderData(chart, tpl, chartDataScripts);
      }
    }
    return new Template.ListReplacer(chartDataScripts);
  }
}
