package org.benchkit.htmlchartsreporter;

import static org.benchkit.htmlchartsreporter.ChartRenderUtil.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.HdrHistogram.Histogram;
import org.HdrHistogram.HistogramData;

public class LatencyHistogramChart implements ChartRender {
  private static class Percentile {
    int tick;
    double percentile;
    String label;
    
    public Percentile(int tick, double percentile, String label) {
      this.tick = tick;
      this.percentile = percentile;
      this.label = label;
    }
    
    public String toString() {
      return "[" + tick + ", '" + label + "']";
    }
  }
  
  private static final List<Percentile> percentiles;
  static {
    String format = "%.5f";
    double ip = 100.0;
    int tick = 1;
    List<Percentile> ps = new ArrayList<Percentile>();
    while (ip > 0.00001) {
      double d = 100.0 - ip;
      ps.add(new Percentile(tick, d, String.format(format, d)));
      ip = ip * 0.5;
      tick++;
    }
    ps.add(new Percentile(tick, 100.0, String.format(format, 100.0)));
    percentiles = Collections.unmodifiableList(ps);
  }
  
  private final String chartName;
  private final String xaxisName;
  private final String yaxisName;
  private final Collection<Comparable<?>> xdataPoints;

  public LatencyHistogramChart(String chartName, String xaxisName) {
    this.chartName = chartName;
    this.xaxisName = xaxisName;
    this.yaxisName = "Latency in Milliseconds";
    xdataPoints = new ArrayList<Comparable<?>>();
  }

  @Override
  public void calibrate(Collection<Chart> charts) {
    Set<Comparable<?>> xs = new TreeSet<Comparable<?>>();
    for (Chart chart : charts) for (DataSeries series : chart)
      for (List<DataPoint> points : series) for (DataPoint point : points) {
        xs.add(point.xvalue);
    }
    xdataPoints.addAll(xs);
  }

  @Override
  public void renderData(Chart data, Template tpl, List<String> outputs) {
    for (Comparable<?> xvalue : xdataPoints) {
      tpl.setReplacer("panel", getIdentifier(data, xvalue));
      tpl.setReplacer("xaxisName", "Percentile");
      tpl.setReplacer("yaxisName", yaxisName);
      
      tpl.setReplacer("xaxisOptions",
          ", ticks: " + JsonOut.json(percentiles));
      tpl.setReplacer("yaxisOptions",
          ", transform: function (v) { return Math.log(v+0.0001); }" +
          ", ticks: function (axis) {" +
          " var r=[], n=0.0001, m=axis.max;" +
          " do {" +
          "   n = n * 10;" +
          "   r.push([n, n.toFixed(3)]);" +
          " } while (n < m);" +
          " return r;" +
          "}" +
          ", tickDecimals: 3");
      
      List<Object> dataset = new ArrayList<Object>();
      renderDataset(xvalue, data, dataset);
      tpl.setReplacer("dataset", JsonOut.json(dataset));
      outputs.add(tpl.toString());
    }
  }

  @SuppressWarnings("unchecked")
  private void renderDataset(
      Comparable<?> xvalue,
      Chart data,
      List<Object> dataset) {
    double millisecFactor = 0.0;
    for (DataSeries series : data) {
      List<List<?>> dataPoints = new ArrayList<List<?>>();
      Histogram histogram = null;
      for (List<DataPoint> points : series) {
        for (DataPoint point : points) {
          if (point.xvalue.compareTo(xvalue) == 0) {
            if (histogram == null) {
              histogram = point.data.getHisogram().copy();
              millisecFactor = point.factorConvertingLatencyToMillis;
            } else {
              histogram.add(point.data.getHisogram());
            }
          }
        }
      }
      
      if (histogram == null) {
        continue;
      }
      
      HistogramData histdata = histogram.getHistogramData();
      for (Percentile percentile : percentiles) {
        double value = histdata.getValueAtPercentile(percentile.percentile);
        dataPoints.add(Arrays.asList(percentile.tick, value / millisecFactor));
      }
      
      Hsl hsl = series.getSeriesColour();
      Map<String, Object> linesConfig = createLinesConfig();
      Map<String, Object> curve = createCurve(linesConfig, dataPoints, hsl);
      label(series.getSeriesName()).decorate(curve, linesConfig);
      
      dataset.add(curve);
    }
  }

  @Override
  public void renderPanels(Chart chart, Template panelTpl, Template headlineTpl, List<String> divs) {
    headlineTpl.setReplacer("chartName", chartName);
    divs.add(headlineTpl.toString());
    for (Comparable<?> xvalue : xdataPoints) {
      panelTpl.setReplacer("chartType", "Latency Histogram [" + xaxisName + "=" + xvalue + "]");
      panelTpl.setReplacer("chartId", getIdentifier(chart, xvalue));
      panelTpl.setReplacer("panelClasses", "panel vert-x-labels");
      divs.add(panelTpl.toString());
    }
  }

  private String getIdentifier(Chart chart, Comparable<?> obj) {
    return "latencyhist-chart-" + obj + "-" + chart.getId();
  }
}
