package org.benchkit.htmlchartsreporter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class Template {
  private final List<Object> parts;
  private final Map<String, List<Integer>> replacers;
  private int sizeHint;

  public Template(String resource) {
    parts = new ArrayList<Object>();
    Map<String, List<Integer>> replacers = new HashMap<String, List<Integer>>();
    parseInto(resource, parts, replacers);
    this.replacers = Collections.unmodifiableMap(replacers);
    this.sizeHint = resource.length();
  }
  
  private Template(List<Object> parts, Map<String, List<Integer>> replacers) {
    this.parts = parts;
    this.replacers = replacers;
  }
  
  public Template copy() {
    ArrayList<Object> partsCopy = new ArrayList<Object>(parts);
    HashMap<String, List<Integer>> replacersCopy =
        new HashMap<String, List<Integer>>(replacers);
    return new Template(partsCopy, replacersCopy);
  }

  private static void parseInto(
      String resource,
      List<Object> parts,
      Map<String, List<Integer>> replacers) {
    int i = 0;
    int len = resource.length();
    int mark = 0;
    boolean readingReplacer = false;
    
    while (i < len) {
      char ch = resource.charAt(i);
      if (readingReplacer && ch == '}') {
        String replacerName = resource.substring(mark, i);
        String replacer = resource.substring(mark - 1, i + 1);
        add(replacers, replacerName, parts.size());
        parts.add(replacer);
        readingReplacer = false;
        mark = i + 1;
      } else if (readingReplacer && ch == '{') {
        String part = resource.substring(mark - 1, i);
        parts.add(part);
        mark = i + 1;
      } else if (!readingReplacer && ch == '{') {
        String part = resource.substring(mark, i);
        parts.add(part);
        readingReplacer = true;
        mark = i + 1;
      }
      i++;
    }
    String part = resource.substring(mark, i);
    parts.add(part);
  }

  private static void add(
      Map<String, List<Integer>> replacers,
      String replacerName,
      int index) {
    List<Integer> indices = replacers.get(replacerName);
    if (indices == null) {
      indices = new ArrayList<Integer>();
      replacers.put(replacerName, indices);
    }
    indices.add(index);
  }

  public Template setReplacer(String replacerName, Object replacementValue) {
    List<Integer> replacerIndices = replacers.get(replacerName);
    if (replacerIndices == null) {
      String msg = "No such replacer: '" + replacerName + "'";
      throw new IllegalArgumentException(msg);
    }
    for (int index : replacerIndices) {
      parts.set(index, replacementValue);
    }
    return this;
  }
  
  public String toString() {
    StringBuilder sb = new StringBuilder(sizeHint);
    for (Object part : parts) {
      sb.append(part);
    }
    String result = sb.toString();
    sizeHint = result.length();
    return result;
  }
  
  public static class ListReplacer {
    private final Collection<?> coll;
    
    public ListReplacer(Collection<?> coll) {
      this.coll = coll;
    }
    
    public String toString() {
      StringBuilder sb = new StringBuilder();
      for (Object obj : coll) {
        sb.append(obj);
      }
      return sb.toString();
    }
  }
}
