package org.benchkit.htmlchartsreporter;

import static org.benchkit.htmlchartsreporter.ChartRenderUtil.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.benchkit.htmlchartsreporter.ChartRenderUtil.CurveDecorator;

public class ThroughputChart implements ChartRender {
  private final String chartName;
  private final String xaxisName;
  private final String yaxisName;
  private Map<Comparable<?>, Integer> xaxisTickMap;

  public ThroughputChart(String chartName, String xaxisName, String yaxisName) {
    this.chartName = chartName;
    this.xaxisName = xaxisName;
    this.yaxisName = yaxisName;
  }

  @Override
  public void calibrate(Collection<Chart> charts) {
    Set<Comparable<?>> xaxis = new TreeSet<Comparable<?>>();
    for (Chart chart : charts) for (DataSeries series : chart)
      for (List<DataPoint> points : series) for (DataPoint point : points)
        xaxis.add(point.xvalue);
    int tick = 1;
    Map<Comparable<?>, Integer> xaxisTicks = new TreeMap<Comparable<?>, Integer>();
    for (Comparable<?> value : xaxis) {
      xaxisTicks.put(value, tick);
      tick++;
    }
    xaxisTickMap = Collections.unmodifiableMap(xaxisTicks);
  }

  @SuppressWarnings("unchecked")
  @Override
  public void renderData(Chart data, Template tpl, List<String> outputs) {
    String panelName = getIdentifier(data);
    
    List<Object> dataset = new ArrayList<Object>();
    for (DataSeries series : data) {
      Hsl colour = series.getSeriesColour();
      int id = series.getId();
      String seriesName = series.getSeriesName();
      
      // Render the normal mean throughput curve:
      renderCurve(throughputMean, dataset, series, xaxisTickMap, colour, label(seriesName));
      
      // Render the fill-betweens showing the scatter, if any:
      int curves = series.countPotentialCurves();
      if (curves > 1) {
        series.sortDataPoints(byOpssec);
        for (int i = 0; i < curves; i++) {
          CurveDecorator decorator = withNormalFillBetween(id, colour, i, curves);
          renderCurve(specificOpssec(i), dataset, series, xaxisTickMap, colour, decorator);
        }
      }
    }
    
    List<List<?>> ticks = new ArrayList<List<?>>();
    for (Entry<Comparable<?>, Integer> entry : xaxisTickMap.entrySet()) {
      List<?> tick = Arrays.asList(entry.getValue(), entry.getKey());
      ticks.add(tick);
    }
    tpl.setReplacer("xaxisOptions",
        ", ticks: " + JsonOut.json(ticks));
    tpl.setReplacer("yaxisOptions",
        ", tickFormatter: function (x, axis) {" + 
        "    x += '';" + 
        "    var rgx = /^(\\d+)(\\d{3})(.*)/;" + 
        "    while (rgx.test(x))" + 
        "        x = x.replace(rgx, '$1,$2$3');" + 
        "    return x;" + 
        "}");
    render(tpl, panelName, xaxisName, yaxisName, dataset);
    outputs.add(tpl.toString());
  }

  private String getIdentifier(Chart data) {
    return "throughput-chart-" + data.getId();
  }

  @Override
  public void renderPanels(Chart chart, Template panelTpl, Template headlineTpl, List<String> divs) {
    headlineTpl.setReplacer("chartName", chartName);
    panelTpl.setReplacer("chartType", "Throughput");
    panelTpl.setReplacer("chartId", getIdentifier(chart));
    panelTpl.setReplacer("panelClasses", "panel");
    divs.add(panelTpl.toString());
  }
}
